from typing import TypedDict, Literal, NotRequired, Protocol, Optional, Any
from pandas import DataFrame


class ScikitModel(Protocol):
    def predict_proba(self, X) -> Optional[Any]: ...


class ViewMeta(TypedDict):
    mode: NotRequired[Literal["public", "protected", "private"]]
    view_id: NotRequired[str]
    dataset_name: NotRequired[str]


class ValuateModelOutput(TypedDict):
    view_id: str
    view_url: str


class Credentials(TypedDict):
    email: str
    password: str


class AWSVariables(TypedDict):
    api_url: NotRequired[str]
    app_client_id: NotRequired[str]
    identity_pool_id: NotRequired[str]
    user_pool_id: NotRequired[str]
    bucket_name: NotRequired[str]
    base_url: NotRequired[str]
    validation_api_url: NotRequired[str]


class ColumnNames(TypedDict):
    dataset_column_names: NotRequired[list[str]]
    score_column_names: NotRequired[list[str]]
    dependent_variable_name: NotRequired[str]



from aiohttp import ClientSession
from webbrowser import open_new
from numpy import ndarray
import logging


def launch_browser(base_url: str, view_id: str):
    url = f"{base_url}{view_id}"
    open_new(url)


async def validate_config(validation_url: str, config: dict) -> dict:
    async with ClientSession() as session:
        output = None
        async with session.post(validation_url, json=config) as response:
            if response.status == 200:
                output = await response.json()
            else:
                error = await response.text()
                logging.error(f"Validation failed with {response.status}: {error}")
                raise Exception(f"Validation failed with {response.status}: {error}")
        return output


def get_dataset_column_names(dataset: ndarray, column_names: list[str]) -> list[str]:
    column_count = (
        len(dataset) if dataset.ndim == 1 else len(dataset[0])
    )  # ndim checks the dimensions of numpy array
    # If column names mismatch, log warning
    if len(column_names) != column_count:
        logging.warning(
            f"Mismatch: Dataset has {column_count} columns, but {len(column_names)} column names were provided."
        )

    # Ensure dataset_column_names has enough names, filling missing ones
    output = column_names[:column_count] + [
        f"column-{i+1}" for i in range(len(column_names), column_count)
    ]

    return output


# TODO: Verify if need to handle multi output classifiers.
def get_score_column_names(scores: list[str], column_names: list[str]) -> list[str]:
    column_names_count = len(column_names)
    scores_count = len(scores)

    # Log warning if there's a mismatch
    if column_names_count != scores_count:
        logging.warning(
            f"Mismatch: Scores has {scores_count} possible values, but {column_names_count} column names were provided."
        )

    # Ensure output has enough names, filling missing ones
    output = column_names[:scores_count] + [
        f"score-class-{column_name}" for column_name in column_names[scores_count:]
    ]

    return output



from typing import TypedDict, Literal
from botocore.client import BaseClient
from pandas import DataFrame


class UploadParams(TypedDict):
    s3_client: BaseClient  # Ref: https://stackoverflow.com/a/68006865
    identity_id: str
    file_name: str
    mode: Literal["public", "protected", "private"]
    bucket_name: str
    data: DataFrame | dict


class UploadFileParams(TypedDict):
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    identity_id: str

    data: DataFrame
    file_name: str
    config: dict
    mode: Literal["public", "protected", "private"]
    bucket_name: str



from datetime import datetime
from uuid import uuid4
from gooder_ai import globals
from gooder_ai.s3 import upload_files
from gooder_ai.view import execute_graphql_query, ExecuteGraphQLParams
from gooder_ai.auth import authenticate
from gooder_ai.utils import (
    validate_config,
    launch_browser,
    get_dataset_column_names,
    get_score_column_names,
)
from gooder_ai.types import (
    ViewMeta,
    ValuateModelOutput,
    ScikitModel,
    Credentials,
    AWSVariables,
    ColumnNames,
)
from numpy import ndarray, column_stack
from pandas import DataFrame
import logging


async def valuate_model(
    model: ScikitModel,
    x_data: ndarray,
    y: ndarray,
    auth_credentials: Credentials,
    config: dict,
    view_meta: ViewMeta,
    column_names: ColumnNames = {},
    aws_variables: AWSVariables = {},
) -> ValuateModelOutput:
    logging.info("Model valuation started.")
    email = auth_credentials["email"]
    password = auth_credentials["password"]
    mode = view_meta.get("mode", "private")
    view_id = view_meta.get("view_id", None)
    dataset_name = view_meta.get("dataset_name", f"{datetime.now().timestamp()}")

    # AWS Global Variables
    api_url = aws_variables.get("api_url", globals.API_URL)
    app_client_id = aws_variables.get("app_client_id", globals.App_Client_ID)
    identity_pool_id = aws_variables.get("identity_pool_id", globals.Identity_Pool_ID)
    user_pool_id = aws_variables.get("user_pool_id", globals.User_Pool_ID)
    bucket_name = aws_variables.get("bucket_name", globals.Bucket_Name)
    base_url = aws_variables.get("base_url", globals.Base_URL)
    validation_api_url = aws_variables.get(
        "validation_api_url", globals.Validation_API_URL
    )

    scorer = getattr(
        model, "predict_proba", None
    )  # We are using predict_proba as scorer.
    if scorer is None:
        logging.error(
            "Failed: Input model instance doesn't have a predict_proba method to score the model performance."
        )
        raise Exception(
            "Failed: Input model instance doesn't have a predict_proba method to score the model performance."
        )

    model_classes = getattr(
        model, "classes_", None
    )  # We are using predict_proba as scorer.
    if model_classes is None:
        logging.error(
            "Failed: Input model instance doesn't have classes to classify the target variable"
        )
        raise Exception(
            "Failed: Input model instance doesn't have classes to classify the target variable."
        )

    # Convert to csv
    scores = scorer(x_data)

    dataset_column_names = get_dataset_column_names(
        x_data, column_names.get("dataset_column_names", [])
    )
    score_column_names = get_score_column_names(
        model_classes, column_names.get("score_column_names", [])
    )
    all_column_names = (
        dataset_column_names
        + [column_names.get("dependent_variable_name", "dependent_variable")]
        + score_column_names
    )

    # TODO: Refactor to avoid using dataframe
    dataset = column_stack((x_data, y, scores))
    data = DataFrame(dataset, columns=all_column_names)

    logging.info("Started: Validating config as per the Gooder AI schema.")
    parsed_config = await validate_config(validation_api_url, config)

    if parsed_config["success"] == False:
        logging.error("Failed: Validating config as per the Gooder AI schema.")
        raise Exception("Invalid configuration", parsed_config["error"])
    else:
        logging.info("Success: Validating config as per the Gooder AI schema.")

    logging.info("Started: Authenticating for the Gooder AI platform.")
    credentials = authenticate(
        {
            "email": email,
            "password": password,
            "app_client_id": app_client_id,
            "identity_pool_id": identity_pool_id,
            "user_pool_id": user_pool_id,
        }
    )
    logging.info("Success: Authenticating for the Gooder AI platform.")

    token = credentials["cognito_client_response"]["AuthenticationResult"][
        "AccessToken"
    ]
    aws_access_key_id = credentials["cognito_credentials"]["Credentials"]["AccessKeyId"]
    aws_secret_access_key = credentials["cognito_credentials"]["Credentials"][
        "SecretKey"
    ]
    aws_session_token = credentials["cognito_credentials"]["Credentials"][
        "SessionToken"
    ]
    identity_id = credentials["cognito_credentials"]["IdentityId"]

    parsed_config["data"][
        "datasetID"
    ] = f"{dataset_name}.csv/Sheet1"  # override datasetID of config to match with dataset.

    logging.info("Started: Uploading config and dataset to the Gooder AI platform.")
    path_dictionary = await upload_files(
        {
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "aws_session_token": aws_session_token,
            "identity_id": identity_id,
            "data": data,
            "config": parsed_config["data"],
            "file_name": f"{dataset_name}",
            "mode": mode,
            "bucket_name": bucket_name,
        }
    )
    csv_path = path_dictionary["csv_path"]
    config_path = path_dictionary["config_path"]

    if csv_path is None or config_path is None:
        logging.error("Failed: Uploading config and dataset to the Gooder AI platform.")
        raise Exception("Failed to upload files")
    else:
        logging.info("Started: Uploading config and dataset to the Gooder AI platform.")

    mutation_type = (
        "updateSharedView" if isinstance(view_id, str) else "createSharedView"
    )

    view_params: ExecuteGraphQLParams = {
        "api_url": api_url,
        "token": token,
        "mutation": mutation_type,
        "variables": {
            "input": {
                "configPath": config_path,
                "datasetPath": csv_path,
                "id": view_id if isinstance(view_id, str) else f"{uuid4()}",
            }
        },
    }

    view = await execute_graphql_query(view_params)
    id: str = view["data"][mutation_type]["id"]
    message = (
        f"View with ID {id} has been successfully updated using the provided view ID: {view_id}."
        if mutation_type == "updateSharedView"
        else f"A new view has been created successfully. Your view ID is {id}. Please save it for future reference and reuse."
    )
    logging.info(message)
    logging.info("Model valuation can be continued on the Gooder AI platform now.")
    launch_browser(base_url, id)
    return {"view_id": id, "view_url": f"{base_url}{id}"}



import boto3
from pandas import DataFrame
from io import BytesIO
from json import dumps
from asyncio import create_task
from botocore.exceptions import ClientError
from gooder_ai.s3.types import UploadParams, UploadFileParams
from gooder_ai import globals


# Convert dataframe to an in-memory csv sheet Ref: https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_csv.html#pandas.DataFrame.to_csv
# Using buffer reference: https://stackoverflow.com/a/40540730
# Using BytesIO instead of StringIO because, s3 upload file function uses BytesIO
async def transform_dataframe_to_csv_file_object(data: DataFrame) -> BytesIO:
    csv_buffer = BytesIO()
    data.to_csv(csv_buffer, index=False)
    csv_buffer.seek(
        0
    )  # When we write something in buffer, file pointer is moved at the end of file. Here we are resetting the file pointer position to start after writing to file. Ref: https://stackoverflow.com/a/53485819
    return csv_buffer


# Convert config dictionary to a string to an in-memory json Ref: https://docs.python.org/3/library/json.html#basic-usage
# Using buffer reference: https://docs.python.org/3/library/io.html#binary-i-o
# Using BytesIO instead of StringIO because, s3 upload file function uses BytesIO
async def transform_config_to_json_file_object(config: dict) -> BytesIO:
    json_string = dumps(config)
    json_buffer = BytesIO(json_string.encode("utf-8"))
    json_buffer.seek(
        0
    )  # When we write something in buffer, file pointer is moved at the end of file. Here we are resetting the file pointer position to start after writing to file. Ref: https://stackoverflow.com/a/53485819
    return json_buffer


async def upload_handler(input: UploadParams) -> str | None:
    s3_client = input["s3_client"]
    identity_id = input["identity_id"]
    data = input["data"]
    mode = input["mode"]
    bucket_name = input["bucket_name"]
    file_name = input["file_name"]

    path: str | None = None
    parsed_data: BytesIO | None = None

    if isinstance(data, DataFrame):
        # The path to upload file is referenced from amplify. Ref: https://docs.amplify.aws/gen1/javascript/build-a-backend/storage/path/#using-protected-accesslevel
        path = f"{mode}/{identity_id}/data/{file_name}.csv"
        parsed_data = await transform_dataframe_to_csv_file_object(data)
    else:
        path = f"{mode}/{identity_id}/config/{file_name}.json"
        parsed_data = await transform_config_to_json_file_object(data)

    try:
        s3_client.upload_fileobj(parsed_data, bucket_name, path)
    except ClientError as e:
        print(f"Error uploading file: {e}")
        path = None
    return path


async def upload_files(input: UploadFileParams) -> dict[str, str | None]:
    aws_access_key_id = input["aws_access_key_id"]
    aws_secret_access_key = input["aws_secret_access_key"]
    aws_session_token = input["aws_session_token"]
    identity_id = input["identity_id"]
    data = input["data"]
    file_name = input["file_name"]
    config = input["config"]
    mode = input["mode"]

    bucket_name = input["bucket_name"]

    # Reference for using temporary credentials : https://docs.aws.amazon.com/IAM/latest/UserGuide/id_credentials_temp_use-resources.html
    s3_client = boto3.client(
        "s3",
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
    )
    # We use async keyword to define a co-routine if we need to execute a co-routine in the background we need to create a task of it.
    # Here we need both of these upload file to run in background that's why we are creating a tasks of them.
    # Ref: https://stackoverflow.com/questions/57966935/asyncio-task-vs-coroutine
    upload_csv_task = create_task(
        upload_handler(
            {
                "identity_id": identity_id,
                "s3_client": s3_client,
                "data": data,
                "file_name": file_name,
                "mode": mode,
                "bucket_name": bucket_name,
            }
        )
    )
    upload_config_task = create_task(
        upload_handler(
            {
                "bucket_name": bucket_name,
                "file_name": file_name,
                "data": config,
                "identity_id": identity_id,
                "mode": mode,
                "s3_client": s3_client,
            }
        )
    )

    csv_path = await upload_csv_task
    config_path = await upload_config_task

    output = {"csv_path": csv_path, "config_path": config_path}
    return output
